# js dec2024 cleaned up code with vulture,
#    updated library calls for OSMnx version 2.0.0
#    updated library calls for Folium 0.19.2

import osmnx as ox
import networkx as nx
import folium
import pickle
import time, datetime
import os.path
import sys
import webbrowser

import os
import glob


import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)
print('Python version: ',sys.version)

start_time = time.time()

def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)

def graph_length(Q):
    length = ox.stats.edge_length_total(Q) #in meters
    length = length * 0.0006213712
    return length

def get_bounds(Q):
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    tb = edges.total_bounds
    print( tb )
    sw = [tb[1],tb[0]]
    ne = [tb[3],tb[2]]
    return [sw, ne]

def check_node_geometry(Q):
    count = 0
    for node, dat in Q.nodes(data=True):
        if 'x' not in dat:
            count += 1
            #d = Q.degree(node)
            #print('node ', node, ' has no geometry attribute, degree ', d)t
    if count>0:
        print('number of nodes without geometry: ',count)


def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    fours = 0
    four_pairs = 0
    parallels = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
        if d==4:
            fours += 1
            for neighbor in Q.neighbors(node):
                if Q.degree(neighbor)==4 and neighbor>node: # prevent double-count
                    four_pairs += 1
        if d>6:
            print('node ', node, 'has degree ', d)
        for neighbor in Q.neighbors(node):
            edge_count = Q.number_of_edges(node,neighbor)
            if edge_count > 1:
                parallels += edge_count - 1
                
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of 4-way intersections ' + txt + ' :', fours)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('number of parallel edges ' + txt + ' :', parallels)
    if four_pairs > 0:
        print('number of 4-way adjacent pairs ' + txt + ' :', four_pairs)

        
    check_node_geometry(Q)
    print('')

def colorize_graph(Q):
    QR = Q.copy()
    for node, dat in Q.nodes(data=True):
        d = Q.degree(node)
        dat['color'] = d

        QR.add_node(node,**dat)                #replace node value

    for u, v, k, dat in Q.edges(keys=True,data=True):
        d = 'path'
        if 'highway' in dat:
            d = dat['highway']
        dat['color']=0.0
        if isinstance(d, str):
            if d=='path' or d == 'footway':
                dat['color']=1.0
            else:
                dat['color'] = 2.0
        QR.add_edge(u,v,key=k,**dat)
    QR.graph['color']=True
    return QR

town_markers = [(34.29963, -108.13268, 'Pie Town'),
         (33.19593, -108.20834, 'Gila Hot Springs'),
         (32.77981, -108.27533, 'Silver City'),
         (33.37939, -108.90369, 'Alma/Glenwood'),
         (33.03022, -108.16864, 'Lake Roberts'),
         (32.85594, -107.98024, 'Mimbres'),
         (32.91685, -107.70593, 'Kingston')]

# draw() changed after moving to Linux
def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    T = Q.copy()
    if not 'color' in T.graph:
        T = colorize_graph(T)
    QD = T.to_directed()
    QD.graph['crs'] = ox.settings.default_crs

    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, remove_rings=True, track_merged=True)
    
    nodes, edges = ox.graph_to_gdfs(QD)
#    m = edges.explore(color='#ff00ff', tiles="openstreetmap",tooltip='name')
    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap",tooltip='name')
    if show_nodes:    
#        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    
        nodes.explore(m=m, column='color', cmap='tab10', legend=False,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)

def save_png(Q, max_Q=None, bounds=None, index = 0):
    m = folium.Map()
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    length = graph_length(Q)
    iteration_label = 'iteration '+ str(index) + ' ' + str(round(length)) + ' miles'
    ax = edges.plot(color='#0000ff', alpha=0.2, zorder=1, label = iteration_label)

    if max_Q != None:
        QD = max_Q.to_directed()
        QD.graph['crs'] = ox.settings.default_crs
        nodes, edges = ox.graph_to_gdfs(QD)
        max_length = graph_length(max_Q)
        max_label = 'current max: ' + str(round(max_length)) + ' miles'            
        edges.plot(ax=ax, color='#ff0000', alpha=1.0 , zorder=0, linestyle='dotted', label = max_label)
    if bounds != None:
        sw = bounds[0]
        ne = bounds[1]
        ax.set_xlim(sw[1], ne[1])
        ax.set_ylim(sw[0], ne[0])
    ax.legend(loc="lower left")
    ax.plot()
    plt.axis('off')
    plt.savefig('fig/img{}.png'.format(index), dpi=120)
    plt.close()    

H = pickle.load(open('h_with_roadwalks_v2.pkl', 'rb'))
report(H,'graph loaded from disk')
#Note to self: We cannot draw until graph is simplified, unknown GDF problem
# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(H), key=len)
J = nx.subgraph(H, nodes).copy()
J = J.to_undirected()
report(J,'in largest subgraph')

T = J.to_directed()
T = ox.simplification.simplify_graph(T, remove_rings=True, track_merged=True)
T = T.to_undirected()
#draw(T,'before deleting some trails')


#delete png files from any previous run
files = glob.glob('fig/' + 'img*.png')
for f in files:
    os.remove(f)
bounds = get_bounds(T)
save_png(T,None,bounds,0)   

not_recommended = ['deloche and winn canyon', #179
                   'catwalk trail', #207 bummer
                   'redstone', #206
                   'spider creek',
                   'little dry', #180
                   'west fork mogollon', #224
                   'trail canyon', #169
                   '#158', #158
                  # 'ring canyon', #162
                   'iron creek', #172
                   'golden link', #218
                   #'#153', #153 mogollon creek trail  only a little section bad
                   '#217', #217
                   '#709', #709 east fork private property
 
                   ######## aldo side
                   'aspen mountain', #814
                   '#73', #73
                   # keep this one for now 'black range crest', #79
                   'middle fork mimbres', #78
                   'sids prong', #121
                   '#120', #120                   
                   ]

edges_to_remove = []
trails_removed = set()
for u,v,k,d in J.edges(keys=True,data=True):
    if 'name' in d:
        trail_name = d['name']
        trail_name = trail_name.lower()
        for nr in not_recommended:
            if nr in trail_name:
#        if any(e for e in not_recommended if e in trail_name):
                edges_to_remove.append( (u,v,k) )
                trails_removed.add(trail_name)
                print('removed: ', trail_name, ' from ', nr)
J.remove_edges_from(edges_to_remove)

print('trails removed: ', trails_removed)
    


report(J,'not-recommended trails removed')

T = J.to_directed()
T = ox.simplification.simplify_graph(T, remove_rings=True, track_merged=True)
T = T.to_undirected()
draw(T,'not-recommended trails removed', show_nodes=True)
save_png(T,None,bounds,1)   


pickle.dump(J,open('j_with_roadwalks_pruned.pkl', 'wb'))

elapsed_time(start_time)
print('finished script ' + os.path.basename(__file__))
